@@ -6,13 +6,13 @@
 ##  This script is derived from mkcert.sh from the mod_ssl distribution.
 ##  It's based on three external programs:
 ##
-##  ssleay ... get it from ftp://ftp.psy.uq.oz.au/pub/Crypto/SSL/ 
+##  openssl .. get it from http://www.openssl.org/
 ##  ca-fix ... get it from http://www.drh-consultancy.demon.co.uk/ca-fix.html
 ##  pkcs12 ... get it from http://www.drh-consultancy.demon.co.uk/pkcs12faq.html
 ##
 
 #   parameters
-ssleay="/sw/pkg/ssleay/bin/ssleay"
+openssl="/e/openssl/pkg/openssl-0.9.2b/bin/openssl"
 cafix="/sw/pkg/ssleay/bin/ca-fix"
 pkcs12="/sw/pkg/ssleay/bin/pkcs12"
 sslcrtdir="."
@@ -61,12 +61,12 @@ if [ ! -f $sslcrtdir/ca.crt ]; then
         touch $HOME/.rnd
     fi
     if [ ".$randfiles" != . ]; then
-        $ssleay genrsa -rand $randfiles \
-                       -out $sslkeydir/ca.key \
-                       1024
+        $openssl genrsa -rand $randfiles \
+                        -out $sslkeydir/ca.key \
+                        1024
     else
-        $ssleay genrsa -out $sslkeydir/ca.key \
-                       1024
+        $openssl genrsa -out $sslkeydir/ca.key \
+                        1024
     fi
     if [ $? -ne 0 ]; then
         echo "mkcert.sh:Error: Failed to generate RSA private key" 1>&2
@@ -99,10 +99,10 @@ emailAddress                    = "7. Email Address            (eg, name@FQDN)"
 emailAddress_max                = 40
 emailAddress_default            = ca@snakeoil.dom
 EOT
-    $ssleay req -config .mkcert.cfg \
-                -new \
-                -key $sslkeydir/ca.key \
-                -out $sslcsrdir/ca.csr
+    $openssl req -config .mkcert.cfg \
+                 -new \
+                 -key $sslkeydir/ca.key \
+                 -out $sslcsrdir/ca.csr
     if [ $? -ne 0 ]; then
         echo "mkcert.sh:Error: Failed to generate certificate signing request" 1>&2
         exit 1
@@ -110,11 +110,11 @@ EOT
     echo "______________________________________________________________________"
     echo ""
     echo "${T_MD}STEP 3: Generating X.509 certificate for CA signed by itself${T_ME}"
-    $ssleay x509 -req \
-                 -days 365 \
-                 -signkey $sslkeydir/ca.key \
-                 -in  $sslcsrdir/ca.csr \
-                 -out $sslcrtdir/ca.crt
+    $openssl x509 -req \
+                  -days 365 \
+                  -signkey $sslkeydir/ca.key \
+                  -in  $sslcsrdir/ca.csr \
+                  -out $sslcrtdir/ca.crt
     if [ $? -ne 0 ]; then
         echo "mkcert.sh:Error: Failed to generate self-signed CA certificate" 1>&2
         exit 1
@@ -136,7 +136,7 @@ EOT
     echo "______________________________________________________________________"
     echo ""
     echo "${T_MD}RESULT:${T_ME}"
-    $ssleay verify $sslcrtdir/ca.crt
+    $openssl verify $sslcrtdir/ca.crt
     if [ $? -ne 0 ]; then
         echo "mkcert.sh:Error: Failed to verify resulting X.509 certificate" 1>&2
         exit 1
@@ -153,12 +153,12 @@ if [ ! -f $HOME/.rnd ]; then
     touch $HOME/.rnd
 fi
 if [ ".$randfiles" != . ]; then
-    $ssleay genrsa -rand $randfiles \
-                   -out $sslkeydir/user.key \
-                   1024
+    $openssl genrsa -rand $randfiles \
+                    -out $sslkeydir/user.key \
+                    1024
 else
-    $ssleay genrsa -out $sslkeydir/user.key \
-                   1024
+    $openssl genrsa -out $sslkeydir/user.key \
+                    1024
 fi
 if [ $? -ne 0 ]; then
     echo "mkcert.sh:Error: Failed to generate RSA private key" 1>&2
@@ -191,10 +191,10 @@ emailAddress                    = "7. Email Address            (eg, name@fqdn)"
 emailAddress_max                = 40
 emailAddress_default            = www@snakeoil.dom
 EOT
-$ssleay req -config .mkcert.cfg \
-            -new \
-            -key $sslkeydir/user.key \
-            -out $sslcsrdir/user.csr
+$openssl req -config .mkcert.cfg \
+             -new \
+             -key $sslkeydir/user.key \
+             -out $sslcsrdir/user.csr
 if [ $? -ne 0 ]; then
     echo "mkcert.sh:Error: Failed to generate certificate signing request" 1>&2
     exit 1
@@ -205,12 +205,12 @@ echo "${T_MD}STEP 7: Generating X.509 certificate signed by own CA${T_ME}"
 if [ ! -f .mkcert.serial ]; then
     echo '01' >.mkcert.serial
 fi
-$ssleay x509 -days 365 \
-             -CAserial .mkcert.serial \
-             -CA $sslcrtdir/ca.crt \
-             -CAkey $sslkeydir/ca.key \
-             -in $sslcsrdir/user.csr -req \
-             -out $sslcrtdir/user.crt
+$openssl x509 -days 365 \
+              -CAserial .mkcert.serial \
+              -CA $sslcrtdir/ca.crt \
+              -CAkey $sslkeydir/ca.key \
+              -in $sslcsrdir/user.csr -req \
+              -out $sslcrtdir/user.crt
 if [ $? -ne 0 ]; then
     echo "mkcert.sh:Error: Failed to generate X.509 certificate" 1>&2
     exit 1
@@ -236,9 +236,9 @@ if [ $? -ne 0 ]; then
     exit 1
 fi
 rm -f $sslcrtdir/user-tmp.crt
-caname="`$ssleay x509 -noout -text -in $sslcrtdir/ca.crt |\
+caname="`$openssl x509 -noout -text -in $sslcrtdir/ca.crt |\
          grep Subject: | sed -e 's;.*CN=;;' -e 's;/Em.*;;'`"
-username="`$ssleay x509 -noout -text -in $sslcrtdir/user.crt |\
+username="`$openssl x509 -noout -text -in $sslcrtdir/user.crt |\
             grep Subject: | sed -e 's;.*CN=;;' -e 's;/Em.*;;'`"
 $pkcs12 \
     -export \
@@ -251,7 +251,7 @@ $pkcs12 \
 echo "______________________________________________________________________"
 echo ""
 echo "${T_MD}RESULT:${T_ME}"
-$ssleay verify -CAfile $sslcrtdir/ca.crt $sslcrtdir/user.crt
+$openssl verify -CAfile $sslcrtdir/ca.crt $sslcrtdir/user.crt
 if [ $? -ne 0 ]; then
     echo "mkcert.sh:Error: Failed to verify resulting X.509 certificate" 1>&2
     exit 1
@@ -260,7 +260,7 @@ $cafix -in $sslcrtdir/user.crt -inkey $sslkeydir/user.key -print
 echo "______________________________________________________________________"
 echo ""
 echo "${T_MD}STEP 8: Enrypting RSA private key of USER with a pass phrase for security${T_ME}"
-$ssleay rsa -des3 -in $sslkeydir/user.key -out $sslkeydir/user.key.crypt
+$openssl rsa -des3 -in $sslkeydir/user.key -out $sslkeydir/user.key.crypt
 if [ $? -ne 0 ]; then
     echo "mkcert.sh:Error: Failed to encrypt RSA private key" 1>&2
     exit 1
