@@ -1,8 +1,8 @@
 /*                      _             _
-**  _ __ ___   ___   __| |    ___ ___| |
-** | '_ ` _ \ / _ \ / _` |   / __/ __| |
-** | | | | | | (_) | (_| |   \__ \__ \ | mod_ssl - Apache Interface to OpenSSL
-** |_| |_| |_|\___/ \__,_|___|___/___/_| http://www.engelschall.com/sw/mod_ssl/
+**  _ __ ___   ___   __| |    ___ ___| |  mod_ssl
+** | '_ ` _ \ / _ \ / _` |   / __/ __| |  Apache Interface to OpenSSL
+** | | | | | | (_) | (_| |   \__ \__ \ |  www.modssl.org
+** |_| |_| |_|\___/ \__,_|___|___/___/_|  ftp.modssl.org
 **                      |_____|
 **  ssl_engine_kernel.c
 **  The SSL engine kernel
@@ -27,7 +27,7 @@
  *    software must display the following acknowledgment:
  *    "This product includes software developed by
  *     Ralf S. Engelschall <rse@engelschall.com> for use in the
- *     mod_ssl project (http://www.engelschall.com/sw/mod_ssl/)."
+ *     mod_ssl project (http://www.modssl.org/)."
  *
  * 4. The names "mod_ssl" must not be used to endorse or promote
  *    products derived from this software without prior written
@@ -42,7 +42,7 @@
  *    acknowledgment:
  *    "This product includes software developed by
  *     Ralf S. Engelschall <rse@engelschall.com> for use in the
- *     mod_ssl project (http://www.engelschall.com/sw/mod_ssl/)."
+ *     mod_ssl project (http://www.modssl.org/)."
  *
  * THIS SOFTWARE IS PROVIDED BY RALF S. ENGELSCHALL ``AS IS'' AND ANY
  * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
@@ -174,11 +174,16 @@ void ssl_hook_NewConnection(conn_rec *conn)
      * attach this to the socket. Additionally we register this attachment
      * so we can detach later.
      */
-    ssl = SSL_new(sc->pSSLCtx);
+    if ((ssl = SSL_new(sc->pSSLCtx)) == NULL) {
+        ssl_log(conn->server, SSL_LOG_ERROR|SSL_ADD_SSLERR,
+                "Unable to create a new SSL connection from the SSL context");
+        ap_ctx_set(fb->ctx, "ssl", NULL);
+        ap_bsetflag(fb, B_EOF|B_EOUT, 1);
+        conn->aborted = 1;
+        return;
+    }
     SSL_clear(ssl);
-#if SSL_LIBRARY_VERSION >= 0x0922
     SSL_set_session_id_context(ssl, cpVHostID, strlen(cpVHostID));
-#endif
     SSL_set_app_data(ssl, conn);  /* conn_rec    (available now)     */
     SSL_set_app_data2(ssl, NULL); /* request_rec (available later)   */
     SSL_set_fd(ssl, fb->fd);
@@ -187,35 +192,58 @@ void ssl_hook_NewConnection(conn_rec *conn)
     /*
      *  Configure callbacks for SSL connection
      */
-#if SSL_LIBRARY_VERSION >= 0x0920
     SSL_set_tmp_rsa_callback(ssl, ssl_callback_TmpRSA);
-#endif
+    SSL_set_tmp_dh_callback(ssl,  ssl_callback_TmpDH);
     if (sc->nLogLevel >= SSL_LOG_DEBUG) {
         BIO_set_callback(SSL_get_rbio(ssl), ssl_io_data_cb);
         BIO_set_callback_arg(SSL_get_rbio(ssl), ssl);
     }
 
     /*
-     * Configure the server certificate and private key
-     * which should be used for this connection.
+	 * Configure the server certificate and private key which should be used
+	 * for this connection. Under combined RSA+DSA situation there can be two
+	 * certs for a server. The SSL library automatically handles this
+	 * internally.
      */
-    if (SSL_use_certificate(ssl, sc->px509Certificate) <= 0) {
-        ssl_log(conn->server, SSL_LOG_ERROR|SSL_ADD_SSLERR,
-                "Unable to configure server certificate for connection");
-        SSL_free(ssl);
-        ap_ctx_set(fb->ctx, "ssl", NULL);
-        ap_bsetflag(fb, B_EOF|B_EOUT, 1);
-        conn->aborted = 1;
-        return;
+    if (sc->pPublicCert[SSL_AIDX_RSA] != NULL) {
+        if (SSL_use_certificate(ssl, sc->pPublicCert[SSL_AIDX_RSA]) <= 0) {
+            ssl_log(conn->server, SSL_LOG_ERROR|SSL_ADD_SSLERR,
+                    "Unable to configure RSA server certificate for connection");
+            SSL_free(ssl);
+            ap_ctx_set(fb->ctx, "ssl", NULL);
+            ap_bsetflag(fb, B_EOF|B_EOUT, 1);
+            conn->aborted = 1;
+            return;
+        }
+        if (SSL_use_PrivateKey(ssl, sc->pPrivateKey[SSL_AIDX_RSA]) <= 0) {
+            ssl_log(conn->server, SSL_LOG_ERROR|SSL_ADD_SSLERR,
+                    "Unable to configure server RSA private key for connection");
+            SSL_free(ssl);
+            ap_ctx_set(fb->ctx, "ssl", NULL);
+            ap_bsetflag(fb, B_EOF|B_EOUT, 1);
+            conn->aborted = 1;
+            return;
+        }
     }
-    if (SSL_use_RSAPrivateKey(ssl, sc->prsaKey) <= 0) {
-        ssl_log(conn->server, SSL_LOG_ERROR|SSL_ADD_SSLERR,
-                "Unable to configure server private key for connection");
-        SSL_free(ssl);
-        ap_ctx_set(fb->ctx, "ssl", NULL);
-        ap_bsetflag(fb, B_EOF|B_EOUT, 1);
-        conn->aborted = 1;
-        return;
+    if (sc->pPublicCert[SSL_AIDX_DSA] != NULL) {
+        if (SSL_use_certificate(ssl, sc->pPublicCert[SSL_AIDX_DSA]) <= 0) {
+            ssl_log(conn->server, SSL_LOG_ERROR|SSL_ADD_SSLERR,
+                    "Unable to configure DSA server certificate for connection");
+            SSL_free(ssl);
+            ap_ctx_set(fb->ctx, "ssl", NULL);
+            ap_bsetflag(fb, B_EOF|B_EOUT, 1);
+            conn->aborted = 1;
+            return;
+        }
+        if (SSL_use_PrivateKey(ssl, sc->pPrivateKey[SSL_AIDX_DSA]) <= 0) {
+            ssl_log(conn->server, SSL_LOG_ERROR|SSL_ADD_SSLERR,
+                    "Unable to configure server DSA private key for connection");
+            SSL_free(ssl);
+            ap_ctx_set(fb->ctx, "ssl", NULL);
+            ap_bsetflag(fb, B_EOF|B_EOUT, 1);
+            conn->aborted = 1;
+            return;
+        }
     }
 
     /*
@@ -224,7 +252,7 @@ void ssl_hook_NewConnection(conn_rec *conn)
     ap_ctx_set(fb->ctx, "ssl::client::dn", NULL);
     ap_ctx_set(fb->ctx, "ssl::verify::error", NULL);
 
-    /* 
+    /*
      * We have to manage a I/O timeout ourself, because Apache
      * does it the first time when reading the request, but we're
      * working some time before this happens.
@@ -299,7 +327,7 @@ void ssl_hook_NewConnection(conn_rec *conn)
                 return;
             }
             else if (ap_ctx_get(ap_global_ctx, "ssl::handshake::timeout") == (void *)TRUE) {
-                ssl_log(srvr, SSL_LOG_ERROR, 
+                ssl_log(srvr, SSL_LOG_ERROR,
                         "SSL handshake timed out (client %s, server %s)",
                         conn->remote_ip, cpVHostID);
                 SSL_set_shutdown(ssl, SSL_RECEIVED_SHUTDOWN);
@@ -414,7 +442,7 @@ void ssl_hook_NewConnection(conn_rec *conn)
 
 #ifdef SSL_VENDOR
     /* Allow vendors to do more things on connection time... */
-    ap_hook_use("ap::mod_ssl::vendor::new_connection", 
+    ap_hook_use("ap::mod_ssl::vendor::new_connection",
                 AP_HOOK_SIG2(void,ptr), AP_HOOK_ALL, conn);
 #endif
 
@@ -449,20 +477,20 @@ void ssl_hook_CloseConnection(conn_rec *conn)
      * the TLSv1 standard into account here:
      *
      * | 7.2.1. Closure alerts
-     * |  
+     * |
      * | The client and the server must share knowledge that the connection is
      * | ending in order to avoid a truncation attack. Either party may
      * | initiate the exchange of closing messages.
-     * |  
+     * |
      * | close_notify
      * |     This message notifies the recipient that the sender will not send
      * |     any more messages on this connection. The session becomes
      * |     unresumable if any connection is terminated without proper
      * |     close_notify messages with level equal to warning.
-     * | 
+     * |
      * | Either party may initiate a close by sending a close_notify alert.
      * | Any data received after a closure alert is ignored.
-     * | 
+     * |
      * | Each party is required to send a close_notify alert before closing
      * | the write side of the connection. It is required that the other party
      * | respond with a close_notify alert of its own and close down the
@@ -476,18 +504,18 @@ void ssl_hook_CloseConnection(conn_rec *conn)
      * 4.x) don't send one, so we would hang.
      */
 
-    /* 
+    /*
      * exchange close notify messages, but allow the user
      * to force the type of handshake via SetEnvIf directive
      */
     if (ap_ctx_get(conn->client->ctx, "ssl::flag::unclean-shutdown") == PTRUE) {
-        /* perform no close notify handshake at all 
+        /* perform no close notify handshake at all
            (violates the SSL/TLS standard!) */
         SSL_set_shutdown(ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
         cpType = "unclean";
     }
     else if (ap_ctx_get(conn->client->ctx, "ssl::flag::accurate-shutdown") == PTRUE) {
-        /* send close notify and wait for clients close notify 
+        /* send close notify and wait for clients close notify
            (standard compliant, but usually causes connection hangs) */
         SSL_set_shutdown(ssl, 0);
         cpType = "accurate";
@@ -505,7 +533,7 @@ void ssl_hook_CloseConnection(conn_rec *conn)
     ap_ctx_set(conn->client->ctx, "ssl", NULL);
 
     /* and finally log the fact that we've closed the connection */
-    ssl_log(conn->server, SSL_LOG_INFO, 
+    ssl_log(conn->server, SSL_LOG_INFO,
             "Connection to child %d closed with %s shutdown (server %s)",
             conn->child_num, cpType, ssl_util_vhostid(conn->pool, conn->server));
     return;
@@ -519,7 +547,7 @@ int ssl_hook_ReadReq(request_rec *r)
     SSL *ssl;
 
     /*
-     * Get the SSL connection structure and perform the 
+     * Get the SSL connection structure and perform the
      * delayed interlinking from SSL back to request_rec
      */
     ssl = ap_ctx_get(r->connection->client->ctx, "ssl");
@@ -554,13 +582,13 @@ int ssl_hook_Translate(request_rec *r)
      * Log information about incoming HTTPS requests
      */
     if (ap_is_initial_req(r))
-        ssl_log(r->server, SSL_LOG_INFO, 
+        ssl_log(r->server, SSL_LOG_INFO,
                 "%s HTTPS request received for child %d (server %s)",
-                r->connection->keepalives <= 0 ? 
+                r->connection->keepalives <= 0 ?
                     "Initial (No.1)" :
-                    ap_psprintf(r->pool, "Subsequent (No.%d)", 
+                    ap_psprintf(r->pool, "Subsequent (No.%d)",
                                 r->connection->keepalives+1),
-                r->connection->child_num, 
+                r->connection->child_num,
                 ssl_util_vhostid(r->pool, r->server));
 
     /*
@@ -609,55 +637,6 @@ int ssl_hook_Handler(request_rec *r)
     return HTTP_BAD_REQUEST;
 }
 
-/*
- *  Auth Handler:
- *  Fake a Basic authentication from the X509 client certificate.
- *
- *  This must be run fairly early on to prevent a real authentication from
- *  occuring, in particular it must be run before anything else that
- *  authenticates a user.  This means that the Module statement for this
- *  module should be LAST in the Configuration file.
- */
-int ssl_hook_Auth(request_rec *r)
-{
-    SSLSrvConfigRec *sc = mySrvConfig(r->server);
-    SSLDirConfigRec *dc = myDirConfig(r);
-    char b1[MAX_STRING_LEN], b2[MAX_STRING_LEN];
-    char *clientdn;
-
-    /*
-     * We decline operation in various situations..
-     */
-    if (!sc->bEnabled)
-        return DECLINED;
-    if (ap_ctx_get(r->connection->client->ctx, "ssl") == NULL)
-        return DECLINED;
-    if (!(dc->nOptions & SSL_OPT_FAKEBASICAUTH))
-        return DECLINED;
-    if (r->connection->user)
-        return DECLINED;
-    if ((clientdn = (char *)ap_ctx_get(r->connection->client->ctx, "ssl::client::dn")) == NULL)
-        return DECLINED;
-
-    /*
-     * Fake a password - which one would be immaterial, as, it seems, an empty
-     * password in the users file would match ALL incoming passwords, if only
-     * we were using the standard crypt library routine. Unfortunately, OpenSSL
-     * "fixes" a "bug" in crypt and thus prevents blank passwords from
-     * working.  (IMHO what they really fix is a bug in the users of the code
-     * - failing to program correctly for shadow passwords).  We need,
-     * therefore, to provide a password. This password can be matched by
-     * adding the string "xxj31ZMTZzkVA" as the password in the user file.
-     * This is just the crypted variant of the word "password" ;-)
-     */
-    ap_snprintf(b1, sizeof(b1), "%s:password", clientdn);
-    ssl_util_uuencode(b2, b1, FALSE);
-    ap_snprintf(b1, sizeof(b1), "Basic %s", b2);
-    ap_table_set(r->headers_in, "Authorization", b1);
-
-    return DECLINED;
-}
-
 /*
  *  Access Handler
  */
@@ -676,12 +655,12 @@ int ssl_hook_Access(request_rec *r)
     BOOL renegotiate; 
 #ifdef SSL_EXPERIMENTAL
     BOOL reconfigured_locations;
-    STACK *skCAList;
+    STACK_OF(X509_NAME) *skCAList;
     char *cpCAPath;
     char *cpCAFile;
 #endif
-    STACK *skCipherOld;
-    STACK *skCipher;
+    STACK_OF(SSL_CIPHER)  *skCipherOld;
+    STACK_OF(SSL_CIPHER)  *skCipher;
     int nVerifyOld;
     int nVerify;
     int n;
@@ -698,6 +677,8 @@ int ssl_hook_Access(request_rec *r)
      */
     if (dc->bSSLRequired && ssl == NULL) {
         ap_log_reason("SSL connection required", r->filename, r);
+        /* remember forbidden access for strict require option */
+        ap_table_setn(r->notes, "ssl-access-forbidden", (void *)1);
         return FORBIDDEN;
     }
 
@@ -736,7 +717,7 @@ int ssl_hook_Access(request_rec *r)
     if (dc->szCipherSuite != NULL) {
         /* remember old cipher suite for comparison */
         if ((skCipherOld = SSL_get_ciphers(ssl)) != NULL)
-            skCipherOld = sk_dup(skCipherOld);
+            skCipherOld = sk_SSL_CIPHER_dup(skCipherOld);
         /* configure new cipher suite */
         if (!SSL_set_cipher_list(ssl, dc->szCipherSuite)) {
             ssl_log(r->server, SSL_LOG_WARN|SSL_ADD_SSLERR,
@@ -749,14 +730,14 @@ int ssl_hook_Access(request_rec *r)
             (skCipherOld != NULL && skCipher == NULL)   )
             renegotiate = TRUE;
         else if (skCipherOld != NULL && skCipher != NULL) {
-            for (n = 0; n < sk_num(skCipher); n++) {
-                if (sk_find(skCipherOld, sk_value(skCipher, n)) < 0) {
+            for (n = 0; n < sk_SSL_CIPHER_num(skCipher); n++) {
+                if (sk_SSL_CIPHER_find(skCipherOld, sk_SSL_CIPHER_value(skCipher, n)) < 0) {
                     renegotiate = TRUE;
                     break;
                 }
             }
-            for (n = 0; n < sk_num(skCipherOld); n++) {
-                if (sk_find(skCipher, sk_value(skCipherOld, n)) < 0) {
+            for (n = 0; n < sk_SSL_CIPHER_num(skCipherOld); n++) {
+                if (sk_SSL_CIPHER_find(skCipher, sk_SSL_CIPHER_value(skCipherOld, n)) < 0) {
                     renegotiate = TRUE;
                     break;
                 }
@@ -764,7 +745,7 @@ int ssl_hook_Access(request_rec *r)
         }
         /* free old cipher suite */
         if (skCipherOld != NULL)
-            sk_free(skCipherOld);
+            sk_SSL_CIPHER_free(skCipherOld);
     }
 
     /*
@@ -918,7 +899,7 @@ int ssl_hook_Access(request_rec *r)
      * Check SSLRequire boolean expressions
      */
     apRequirement = dc->aRequirement;
-    pRequirements  = (ssl_require_t *)apRequirement->elts;
+    pRequirements = (ssl_require_t *)apRequirement->elts;
     for (i = 0; i < apRequirement->nelts; i++) {
         pRequirement = &pRequirements[i];
         ok = ssl_expr_exec(r, pRequirement->mpExpr);
@@ -926,6 +907,8 @@ int ssl_hook_Access(request_rec *r)
             cp = ap_psprintf(r->pool, "Failed to execute SSL requirement expression: %s",
                              ssl_expr_get_error());
             ap_log_reason(cp, r->filename, r);
+            /* remember forbidden access for strict require option */
+            ap_table_setn(r->notes, "ssl-access-forbidden", (void *)1);
             return FORBIDDEN;
         }
         if (ok != 1) {
@@ -936,6 +919,8 @@ int ssl_hook_Access(request_rec *r)
                     "Failed expression: %s", pRequirement->cpExpr);
             ap_log_reason("SSL requirement expression not fulfilled "
                           "(see SSL logfile for more details)", r->filename, r);
+            /* remember forbidden access for strict require option */
+            ap_table_setn(r->notes, "ssl-access-forbidden", (void *)1);
             return FORBIDDEN;
         }
     }
@@ -946,13 +931,85 @@ int ssl_hook_Access(request_rec *r)
      */
     rc = OK;
 #ifdef SSL_VENDOR
-    ap_hook_use("ap::mod_ssl::vendor::access_handler", 
-                AP_HOOK_SIG2(int,ptr), AP_HOOK_DECLINE(DECLINED), 
+    ap_hook_use("ap::mod_ssl::vendor::access_handler",
+                AP_HOOK_SIG2(int,ptr), AP_HOOK_DECLINE(DECLINED),
                 &rc, r);
 #endif
     return rc;
 }
 
+/*
+ *  Auth Handler:
+ *  Fake a Basic authentication from the X509 client certificate.
+ *
+ *  This must be run fairly early on to prevent a real authentication from
+ *  occuring, in particular it must be run before anything else that
+ *  authenticates a user.  This means that the Module statement for this
+ *  module should be LAST in the Configuration file.
+ */
+int ssl_hook_Auth(request_rec *r)
+{
+    SSLSrvConfigRec *sc = mySrvConfig(r->server);
+    SSLDirConfigRec *dc = myDirConfig(r);
+    char b1[MAX_STRING_LEN], b2[MAX_STRING_LEN];
+    char *clientdn;
+
+    /*
+     * Additionally forbid access (again)
+     * when strict require option is used.
+     */
+    if (   (dc->nOptions & SSL_OPT_STRICTREQUIRE)
+        && (ap_table_get(r->notes, "ssl-access-forbidden") != NULL))
+        return FORBIDDEN;
+
+    /*
+     * We decline operation in various situations..
+     */
+    if (!sc->bEnabled)
+        return DECLINED;
+    if (ap_ctx_get(r->connection->client->ctx, "ssl") == NULL)
+        return DECLINED;
+    if (!(dc->nOptions & SSL_OPT_FAKEBASICAUTH))
+        return DECLINED;
+    if (r->connection->user)
+        return DECLINED;
+    if ((clientdn = (char *)ap_ctx_get(r->connection->client->ctx, "ssl::client::dn")) == NULL)
+        return DECLINED;
+
+    /*
+     * Fake a password - which one would be immaterial, as, it seems, an empty
+     * password in the users file would match ALL incoming passwords, if only
+     * we were using the standard crypt library routine. Unfortunately, OpenSSL
+     * "fixes" a "bug" in crypt and thus prevents blank passwords from
+     * working.  (IMHO what they really fix is a bug in the users of the code
+     * - failing to program correctly for shadow passwords).  We need,
+     * therefore, to provide a password. This password can be matched by
+     * adding the string "xxj31ZMTZzkVA" as the password in the user file.
+     * This is just the crypted variant of the word "password" ;-)
+     */
+    ap_snprintf(b1, sizeof(b1), "%s:password", clientdn);
+    ssl_util_uuencode(b2, b1, FALSE);
+    ap_snprintf(b1, sizeof(b1), "Basic %s", b2);
+    ap_table_set(r->headers_in, "Authorization", b1);
+
+    return DECLINED;
+}
+
+int ssl_hook_UserCheck(request_rec *r)
+{
+    SSLDirConfigRec *dc = myDirConfig(r);
+
+    /*
+     * Additionally forbid access (again)
+     * when strict require option is used.
+     */
+    if (   (dc->nOptions & SSL_OPT_STRICTREQUIRE)
+        && (ap_table_get(r->notes, "ssl-access-forbidden") != NULL))
+        return FORBIDDEN;
+
+    return DECLINED;
+}
+
 /*
  *   Fixup Handler
  */
@@ -1009,6 +1066,7 @@ static const char *ssl_hook_Fixup_vars[] = {
     "SSL_SERVER_I_DN_Email",
     "SSL_SERVER_A_KEY",
     "SSL_SERVER_A_SIG",
+    "SSL_SESSION_ID",
     NULL
 };
 
@@ -1073,7 +1131,7 @@ int ssl_hook_Fixup(request_rec *r)
  * The background of this as the TLSv1 standard explains it:
  *
  * | D.1. Temporary RSA keys
- * | 
+ * |
  * |    US Export restrictions limit RSA keys used for encryption to 512
  * |    bits, but do not place any limit on lengths of RSA keys used for
  * |    signing operations. Certificates often need to be larger than 512
@@ -1081,7 +1139,7 @@ int ssl_hook_Fixup(request_rec *r)
  * |    transactions or for applications requiring long-term security. Some
  * |    certificates are also designated signing-only, in which case they
  * |    cannot be used for key exchange.
- * | 
+ * |
  * |    When the public key in the certificate cannot be used for encryption,
  * |    the server signs a temporary RSA key, which is then exchanged. In
  * |    exportable applications, the temporary RSA key should be the maximum
@@ -1091,7 +1149,7 @@ int ssl_hook_Fixup(request_rec *r)
  * |    changed daily or every 500 transactions, and more often if possible.
  * |    Note that while it is acceptable to use the same temporary key for
  * |    multiple transactions, it must be signed each time it is used.
- * | 
+ * |
  * |    RSA key generation is a time-consuming process. In many cases, a
  * |    low-priority process can be assigned the task of key generation.
  * |    Whenever a new key is completed, the existing temporary key can be
@@ -1101,17 +1159,12 @@ int ssl_hook_Fixup(request_rec *r)
  * which we now just handle out on demand....
  *
  */
-#if SSL_LIBRARY_VERSION >= 0x0920
 RSA *ssl_callback_TmpRSA(SSL *pSSL, int nExport, int nKeyLen)
-#else
-RSA *ssl_callback_TmpRSA(SSL *pSSL, int nExport)
-#endif
 {
     SSLModConfigRec *mc = myModConfig();
     RSA *rsa;
 
     rsa = NULL;
-#if SSL_LIBRARY_VERSION >= 0x0920
     if (nExport) {
         /* It's because an export cipher is used */
         if (nKeyLen == 512)
@@ -1126,15 +1179,35 @@ RSA *ssl_callback_TmpRSA(SSL *pSSL, int nExport)
         /* It's because a sign-only certificate situation exists */
         rsa = mc->pRSATmpKey1024;
     }
-#else
-    if (nExport)
-        rsa = mc->pRSATmpKey512;
-    else
-        rsa = mc->pRSATmpKey1024;
-#endif
     return rsa;
 }
 
+/* 
+ * Handle out the already generated DH parameters...
+ */
+DH *ssl_callback_TmpDH(SSL *pSSL, int nExport, int nKeyLen)
+{
+    SSLModConfigRec *mc = myModConfig();
+    DH *dh;
+
+    dh = NULL;
+    if (nExport) {
+        /* It's because an export cipher is used */
+        if (nKeyLen == 512)
+            dh = mc->pDHTmpParam512;
+        else if (nKeyLen == 1024)
+            dh = mc->pDHTmpParam1024;
+        else
+            /* it's too expensive to generate on-the-fly, so keep 1024bit */
+            dh = mc->pDHTmpParam1024;
+    }
+    else {
+        /* It's because a sign-only certificate situation exists */
+        dh = mc->pDHTmpParam1024;
+    }
+    return dh;
+}
+
 /*
  * This OpenSSL callback function is called when OpenSSL
  * does client authentication and verifies the certificate chain.
@@ -1199,6 +1272,14 @@ int ssl_callback_SSLVerify(int ok, X509_STORE_CTX *ctx)
         ok = TRUE;
     }
 
+    /*
+     * Additionally perform CRL-based revocation checks
+     */
+    if (ok) {
+        ok = ssl_callback_SSLVerify_CRL(ok, ctx, s);
+        errnum = X509_STORE_CTX_get_error(ctx);
+    }
+
     /*
      * If we already know it's not ok, log the real reason
      */
@@ -1207,7 +1288,7 @@ int ssl_callback_SSLVerify(int ok, X509_STORE_CTX *ctx)
                 errnum, X509_verify_cert_error_string(errnum));
         ap_ctx_set(conn->client->ctx, "ssl::client::dn", NULL);
         ap_ctx_set(conn->client->ctx, "ssl::verify::error",
-                   X509_verify_cert_error_string(errnum));
+                   (void *)X509_verify_cert_error_string(errnum));
     }
 
     /*
@@ -1215,15 +1296,15 @@ int ssl_callback_SSLVerify(int ok, X509_STORE_CTX *ctx)
      */
     if (dc != NULL && dc->nVerifyDepth != UNSET)
         depth = dc->nVerifyDepth;
-    else 
+    else
         depth = sc->nVerifyDepth;
     if (errdepth > depth) {
         ssl_log(s, SSL_LOG_ERROR,
                 "Certificate Verification: Certificate Chain too long "
-                "(chain has %d certificates, but maximum allowed are only %d)", 
+                "(chain has %d certificates, but maximum allowed are only %d)",
                 errdepth, depth);
         ap_ctx_set(conn->client->ctx, "ssl::verify::error",
-                   X509_verify_cert_error_string(X509_V_ERR_CERT_CHAIN_TOO_LONG));
+                   (void *)X509_verify_cert_error_string(X509_V_ERR_CERT_CHAIN_TOO_LONG));
         ok = FALSE;
     }
 
@@ -1233,6 +1314,162 @@ int ssl_callback_SSLVerify(int ok, X509_STORE_CTX *ctx)
     return (ok);
 }
 
+int ssl_callback_SSLVerify_CRL(
+    int ok, X509_STORE_CTX *ctx, server_rec *s)
+{
+    SSLSrvConfigRec *sc;
+    X509_OBJECT obj;
+    X509_NAME *subject;
+    X509_NAME *issuer;
+    X509 *xs;
+    X509_CRL *crl;
+    X509_REVOKED *revoked;
+    long serial;
+    BIO *bio;
+    int i, n, rc;
+    char *cp;
+    char *cp2;
+
+    /*
+     * Unless a revocation store for CRLs was created we
+     * cannot do any CRL-based verification, of course.
+     */
+    sc = mySrvConfig(s);
+    if (sc->pRevocationStore == NULL)
+        return ok;
+
+    /*
+     * Determine certificate ingredients in advance
+     */
+    xs      = X509_STORE_CTX_get_current_cert(ctx);
+    subject = X509_get_subject_name(xs);
+    issuer  = X509_get_issuer_name(xs);
+
+    /*
+     * OpenSSL provides the general mechanism to deal with CRLs but does not
+     * use them automatically when verifying certificates, so we do it
+     * explicitly here. We will check the CRL for the currently checked
+     * certificate, if there is such a CRL in the store.
+     *
+     * We come through this procedure for each certificate in the certificate
+     * chain, starting with the root-CA's certificate. At each step we've to
+     * both verify the signature on the CRL (to make sure it's a valid CRL)
+     * and it's revocation list (to make sure the current certificate isn't
+     * revoked).  But because to check the signature on the CRL we need the
+     * public key of the issuing CA certificate (which was already processed
+     * one round before), we've a little problem. But we can both solve it and
+     * at the same time optimize the processing by using the following
+     * verification scheme (idea and code snippets borrowed from the GLOBUS
+     * project):
+     *
+     * 1. We'll check the signature of a CRL in each step when we find a CRL
+     *    through the _subject_ name of the current certificate. This CRL
+     *    itself will be needed the first time in the next round, of course.
+     *    But we do the signature processing one round before this where the
+     *    public key of the CA is available.
+     *
+     * 2. We'll check the revocation list of a CRL in each step when
+     *    we find a CRL through the _issuer_ name of the current certificate.
+     *    This CRLs signature was then already verified one round before.
+     *
+     * This verification scheme allows a CA to revoke its own certificate as
+     * well, of course.
+     */
+
+    /*
+     * Try to retrieve a CRL corresponding to the _subject_ of
+     * the current certificate in order to verify it's integrity.
+     */
+    memset((char *)&obj, 0, sizeof(obj));
+    rc = SSL_X509_STORE_lookup(sc->pRevocationStore, X509_LU_CRL, subject, &obj);
+    crl = obj.data.crl;
+    if (rc > 0 && crl != NULL) {
+        /*
+         * Log information about CRL
+         * (A little bit complicated because of ASN.1 and BIOs...)
+         */
+        if (ssl_log_applies(s, SSL_LOG_TRACE)) {
+            bio = BIO_new(BIO_s_mem());
+            BIO_printf(bio, "lastUpdate: ");
+            ASN1_UTCTIME_print(bio, X509_CRL_get_lastUpdate(crl));
+            BIO_printf(bio, ", nextUpdate: ");
+            ASN1_UTCTIME_print(bio, X509_CRL_get_nextUpdate(crl));
+            n = BIO_pending(bio);
+            cp = malloc(n+1);
+            n = BIO_read(bio, cp, n);
+            cp[n] = NUL;
+            BIO_free(bio);
+            cp2 = X509_NAME_oneline(subject, NULL, 0);
+            ssl_log(s, SSL_LOG_TRACE, "CA CRL: Issuer: %s, %s", cp2, cp);
+            free(cp);
+            free(cp2);
+        }
+
+        /*
+         * Verify the signature on this CRL
+         */
+        if (X509_CRL_verify(crl, X509_get_pubkey(xs)) <= 0) {
+            ssl_log(s, SSL_LOG_WARN, "Invalid signature on CRL");
+            X509_STORE_CTX_set_error(ctx, X509_V_ERR_CRL_SIGNATURE_FAILURE);
+            X509_OBJECT_free_contents(&obj);
+            return FALSE;
+        }
+
+        /*
+         * Check date of CRL to make sure it's not expired
+         */
+        i = X509_cmp_current_time(X509_CRL_get_nextUpdate(crl));
+        if (i == 0) {
+            ssl_log(s, SSL_LOG_WARN, "Found CRL has invalid nextUpdate field");
+            X509_STORE_CTX_set_error(ctx, X509_V_ERR_ERROR_IN_CRL_NEXT_UPDATE_FIELD);
+            X509_OBJECT_free_contents(&obj);
+            return FALSE;
+        }
+        if (i < 0) {
+            ssl_log(s, SSL_LOG_WARN,
+                    "Found CRL is expired - "
+                    "revoking all certificates until you get updated CRL");
+            X509_STORE_CTX_set_error(ctx, X509_V_ERR_CRL_HAS_EXPIRED);
+            X509_OBJECT_free_contents(&obj);
+            return FALSE;
+        }
+        X509_OBJECT_free_contents(&obj);
+    }
+
+    /*
+     * Try to retrieve a CRL corresponding to the _issuer of
+     * the current certificate in order to check for revokation.
+     */
+    memset((char *)&obj, 0, sizeof(obj));
+    rc = SSL_X509_STORE_lookup(sc->pRevocationStore, X509_LU_CRL, issuer, &obj);
+    crl = obj.data.crl;
+    if (rc > 0 && crl != NULL) {
+        /*
+         * Check if the current certificate is revoked by this CRL
+         */
+        n = sk_num(X509_CRL_get_REVOKED(crl));
+        for (i = 0; i < n; i++) {
+            revoked = (X509_REVOKED *)sk_value(X509_CRL_get_REVOKED(crl), i);
+            if (ASN1_INTEGER_cmp(revoked->serialNumber, X509_get_serialNumber(xs)) == 0) {
+
+                serial = ASN1_INTEGER_get(revoked->serialNumber);
+                cp = X509_NAME_oneline(issuer, NULL, 0);
+                ssl_log(s, SSL_LOG_INFO,
+                        "Certificate with serial %ld (0x%lX) "
+                        "revoked per CRL from issuer %s",
+                        serial, serial, cp);
+                free(cp);
+
+                X509_STORE_CTX_set_error(ctx, X509_V_ERR_CERT_REVOKED);
+                X509_OBJECT_free_contents(&obj);
+                return FALSE;
+            }
+        }
+        X509_OBJECT_free_contents(&obj);
+    }
+    return ok;
+}
+
 /*
  *  This callback function is executed by OpenSSL whenever a new SSL_SESSION is
  *  added to the internal OpenSSL session cache. We use this hook to spread the
@@ -1390,27 +1627,27 @@ void ssl_callback_LogTracingState(SSL *ssl, int where, int rc)
         else if (where & SSL_CB_HANDSHAKE_DONE)
             ssl_log(s, SSL_LOG_TRACE, "%s: Handshake: done", SSL_LIBRARY_NAME);
         else if (where & SSL_CB_LOOP)
-            ssl_log(s, SSL_LOG_TRACE, "%s: Loop: %s", 
+            ssl_log(s, SSL_LOG_TRACE, "%s: Loop: %s",
                     SSL_LIBRARY_NAME, SSL_state_string_long(ssl));
         else if (where & SSL_CB_READ)
             ssl_log(s, SSL_LOG_TRACE, "%s: Read: %s",
                     SSL_LIBRARY_NAME, SSL_state_string_long(ssl));
         else if (where & SSL_CB_WRITE)
-            ssl_log(s, SSL_LOG_TRACE, "%s: Write: %s", 
+            ssl_log(s, SSL_LOG_TRACE, "%s: Write: %s",
                     SSL_LIBRARY_NAME, SSL_state_string_long(ssl));
         else if (where & SSL_CB_ALERT) {
             str = (where & SSL_CB_READ) ? "read" : "write";
-            ssl_log(s, SSL_LOG_TRACE, "%s: Alert: %s:%s:%s\n", 
+            ssl_log(s, SSL_LOG_TRACE, "%s: Alert: %s:%s:%s\n",
                     SSL_LIBRARY_NAME, str,
                     SSL_alert_type_string_long(rc),
                     SSL_alert_desc_string_long(rc));
         }
         else if (where & SSL_CB_EXIT) {
             if (rc == 0)
-                ssl_log(s, SSL_LOG_TRACE, "%s: Exit: failed in %s", 
+                ssl_log(s, SSL_LOG_TRACE, "%s: Exit: failed in %s",
                         SSL_LIBRARY_NAME, SSL_state_string_long(ssl));
             else if (rc < 0)
-                ssl_log(s, SSL_LOG_TRACE, "%s: Exit: error in %s", 
+                ssl_log(s, SSL_LOG_TRACE, "%s: Exit: error in %s",
                         SSL_LIBRARY_NAME, SSL_state_string_long(ssl));
         }
     }
